#!/usr/bin/env python

#****************************************************************************************************************************************************
# Copyright (c) 2014 Freescale Semiconductor, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#
#    * Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#
#    * Neither the name of the Freescale Semiconductor, Inc. nor the names of
#      its contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#****************************************************************************************************************************************************

import os
import os.path
import shutil
from FslBuildGen import IOUtil, MakeFileHelper, Util
from FslBuildGen.DataTypes import *
from FslBuildGen.Exceptions import *

class GeneratorAndroid(object):
    def __init__(self, config, packages, platformName):
        super(GeneratorAndroid, self).__init__()
        copyFrom = config.GetDirectPath("Android")
        filesTopCopy =  self.__GetTemplateFilesToCopy(config, copyFrom)
        self.LibTemplate = IOUtil.ReadFile(IOUtil.Join(config.SDKConfigTemplatePath, "Android_lib.mk"))
        for package in packages:
            if not package.ResolvedPlatformNotSupported:
                if package.Type == PackageType.Library:
                    self.__GenerateLibraryBuildFile(config, package, platformName)
                elif package.Type == PackageType.Executable:
                    # For now we generate a library build file for android
                    self.__GenerateLibraryBuildFile(config, package, platformName)
                    self.__UpdateExeBuilder(config, package, platformName, filesTopCopy, copyFrom)

    def __GetTemplateFilesToCopy(self, config, copyFrom):
        skip = len(copyFrom)+1
        files = []
        for file in IOUtil.GetFilePaths(copyFrom, None):
            files.append(file[skip:])
        return files

    def __GenerateLibraryBuildFile(self, config, package, platformName):
        libName = package.ShortName
        files = MakeFileHelper.CreateList(package.ResolvedBuildSourceFiles)
        libraryDendencies = Util.ExtractNonVirtualNames(package.ResolvedDirectDependencies)

        externalLibraryDependencies = Util.ExtractNames(Util.FilterByType(package.ResolvedDirectExternalDependencies, ExternalDependencyType.StaticLib))

        if package.Type == PackageType.Executable:
            # FIX: since adding this library causes a setup method to be removed during optimzation
            # But the android 'executable' support is experimental
           if "FslDemoPlatform" in libraryDendencies: libraryDendencies.remove("FslDemoPlatform")

        libraryNames = externalLibraryDependencies + libraryDendencies
        libraryNames = MakeFileHelper.CreateList(libraryNames)


        firstOnly = [entry for entry in package.ResolvedBuildAllPublicCPPDefines if entry.IsFirstActualUse]
        cppDefines = Util.ExtractNames(firstOnly)
        cppDefineNames = MakeFileHelper.CreateDefineList(cppDefines)

        firstOnly = [entry for entry in package.ResolvedBuildAllCPPDefines if entry.IsFirstActualUse]
        cppLocalDefines = Util.ExtractNames(firstOnly)
        cppLocalDefineNames = MakeFileHelper.CreateDefineList(cppLocalDefines)

        includeDirsPublic = []
        includeDirs = []

        for extDependency in package.ResolvedDirectExternalDependencies:
            if extDependency.Include != None:
                if extDependency.Access == AccessType.Public:
                    includeDirsPublic.append(extDependency.Include)
                    includeDirs.append(extDependency.Include)
                elif extDependency.Access == AccessType.Private:
                    includeDirs.append(extDependency.Include)
                else:
                    print("ERROR: Unknown access type")

        if package.AbsoluteIncludePath != None:
            includeDirsPublic.append('$(LOCAL_PATH)/include')
            includeDirs.append('$(LOCAL_PATH)/include')

        if len(includeDirsPublic) > 0:
            includeDirsPublic = ' ' + MakeFileHelper.CreateList2(includeDirsPublic)
        else:
            includeDirsPublic = ''
        if len(includeDirs) > 0:
            includeDirs = ' ' + MakeFileHelper.CreateList2(includeDirs)
        else:
            includeDirs = ''

        build = self.LibTemplate
        build = build.replace("##PACKAGE_TARGET_NAME##", libName)
        build = build.replace("##PACKAGE_SOURCE_FILES##", files)
        build = build.replace("##PACKAGE_LIBRARY_DEPENDENCIES##", libraryNames)
        build = build.replace("##PACKAGE_PUBLIC_CPP_FLAGS##", cppDefineNames)
        build = build.replace("##PACKAGE_CPP_FLAGS##", cppLocalDefineNames)
        build = build.replace("##PACKAGE_PUBLIC_INCLUDE_DIRS##", includeDirsPublic)
        build = build.replace("##PACKAGE_INCLUDE_DIRS##", includeDirs)

        dstFile = IOUtil.Join(package.AbsolutePath, "Android.mk")
        if not config.DisableWrite:
            IOUtil.WriteFileIfChanged(dstFile, build)

    def __UpdateExeBuilder(self, config, package, platformName, filesToCopy, srcPath):
        if config.DisableWrite:
            return

        dstPath = IOUtil.Join(package.AbsolutePath, "Android")
        for file in filesToCopy:
            dst = IOUtil.Join(dstPath, file)
            src = IOUtil.Join(srcPath, file)
            IOUtil.CopySmallFile(src, dst)
 