#!/usr/bin/env python

#****************************************************************************************************************************************************
# Copyright (c) 2014 Freescale Semiconductor, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#
#    * Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#
#    * Neither the name of the Freescale Semiconductor, Inc. nor the names of
#      its contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#****************************************************************************************************************************************************

from FslBuildGen.DataTypes import *
from FslBuildGen.PackageBuilder import PackageBuilder
from FslBuildGen.DependencyGraph import DependencyGraph
from FslBuildGen.Exceptions import *
from FslBuildGen import IOUtil, Util, PackageConfig
from FslBuildGen.PackagePlatformVariant import PackagePlatformVariant, PackageCPPDefine
from FslBuildGen.XmlStuff import XmlGenFile

class PackageResolver(object):
    def __init__(self, config, platformName, genFiles, markExternalLibFirstUse):
        super(PackageResolver, self).__init__()
        config.LogPrint("- Processing platform: %s" % (platformName))
        # Do the initial dependency checks
        packageBuilder = PackageBuilder(config, platformName, genFiles)
        packages = packageBuilder.AllPackages

        config.LogPrint("  Resolving")
        # FIX:
        # If we resolve the packages using the top level package build order we can do everything a lot smarter
        # than we currently do
        finalResolveOrder = packageBuilder.TopLevelPackage.ResolvedBuildOrder
        #config.LogPrint("Order: " + ", ".join(Util.ExtractNames(finalResolveOrder)))

        self.__ResolvePlatform(platformName, packages)
        self.__ResolveDirectExternalDependencies(platformName, packages)
        self.__ResolveDirectVariants(finalResolveOrder, platformName)
        self.__CheckVariants(config, finalResolveOrder)
        self.__ResolveAllVariants(finalResolveOrder)
        if markExternalLibFirstUse:
            self.__MarkExternalLibFirstUse(packageBuilder.TopLevelPackage)
        # Everything checks out, so we can now start resolving files for the packages
        self.__ResolveBuildIncludeDirs(config, finalResolveOrder)
        self.__ResolveBuildSourceFiles(packages)
        self.__ResolveBuildIncludeFiles(packages)
        self.__ResolveBuildCPPDefines(config, platformName, finalResolveOrder)
        self.__ResolveBuildUsedFeatures(config, platformName, finalResolveOrder)
        self.__ResolveMakeConfig(packages)
        self.Packages = packages

 
    def __ResolveMakeConfig(self, packages):
        for package in packages:
            normalVariants = []
            virtualVariants = []
            for entry in package.ResolvedAllVariantDict.values():
                if entry.Type == VariantType.Normal:
                    normalVariants.append(entry)
                elif entry.Type == VariantType.Virtual:
                    virtualVariants.append(entry)
                else:
                    raise NotImplementedError("unknown variant type");

            normalVariantNames = Util.ExtractNamesAsMakeEnvironmentVariables(normalVariants)
            virtualVariantNames = Util.ExtractNamesAsMakeEnvironmentVariables(virtualVariants)
            if len(normalVariantNames) > 0 or len(virtualVariantNames) > 0:
                normalVariantNames.sort(key=lambda s: s.lower())
                virtualVariantNames.sort(key=lambda s: s.lower())
                strNormal = "_".join(normalVariantNames)
                strVirtual = "_".join(virtualVariantNames)
                package.ResolvedNormalVariantName = "_" + strNormal if len(strNormal) > 0 else ""
                package.ResolvedVirtualVariantName = "_" + strVirtual if len(strVirtual) > 0 else ""
                package.ResolvedVariantName = package.ResolvedNormalVariantName + package.ResolvedVirtualVariantName
                package.ResolvedMakeConfigName = "%s%s" % (package.ResolvedPlatformName, package.ResolvedVariantName)
            else:
                package.ResolvedMakeConfigName = package.ResolvedPlatformName
                package.ResolvedVariantName = ""
                package.ResolvedNormalVariantName = ""
                package.ResolvedVirtualVariantName = ""
            package.ResolvedMakeObjectPath = "obj/%s/$(config)" % (package.ResolvedMakeConfigName)


    def __ResolveBuildSourceFiles(self, packages):
        for package in packages:
            if not package.IsVirtual:
                startIdx = len(package.AbsolutePath) + 1
                files = IOUtil.GetFilePaths(package.AbsoluteSourcePath, (".cpp", ".c"))
                files = [Util.UTF8ToAscii(file[startIdx:].replace("\\", "/")) for file in files]
                files.sort(key=lambda s: s.lower())
                package.ResolvedBuildSourceFiles = files


    def __ResolveBuildIncludeFiles(self, packages):
        for package in packages:
            if not package.IsVirtual:
                startIdx = len(package.AbsolutePath) + 1
                filesPub = IOUtil.GetFilePaths(package.AbsoluteIncludePath, (".hpp", ".h")) if package.AbsoluteIncludePath != None else []
                filesAll = list(filesPub)
                filesAll += IOUtil.GetFilePaths(package.AbsoluteSourcePath, (".hpp", ".h"))
                filesPub = [Util.UTF8ToAscii(file[startIdx:].replace("\\", "/")) for file in filesPub]
                filesAll = [Util.UTF8ToAscii(file[startIdx:].replace("\\", "/")) for file in filesAll]
                filesPub.sort(key=lambda s: s.lower())
                filesAll.sort(key=lambda s: s.lower())
                package.ResolvedBuildPublicIncludeFiles = filesPub
                package.ResolvedBuildAllIncludeFiles = filesAll


    def __ResolveBuildIncludeDirs(self, config, finalResolveOrder):
        for package in finalResolveOrder:
            hasLocalIncludeDir = False
            includeDirs = set()
            publicIncludeDirs = []
            privateIncludeDirs = []

            # First process the include paths present in this package
            if package.AbsoluteIncludePath != None:
                hasLocalIncludeDir = True
            for extDependency in package.ResolvedDirectExternalDependencies:
                if extDependency.Include != None:
                    if extDependency.Include in includeDirs:
                        raise Exception("External include dir defined multiple times: '%s'" % (extDependency.Include));
                    includeDirs.add(extDependency.Include)
                    if extDependency.Access == AccessType.Public:
                        publicIncludeDirs.append(extDependency.Include)
                    else:
                        privateIncludeDirs.append(extDependency.Include)

            # Then pull in the dependencies from the packages we depend upon
            # here we take advantage of the fact that all packages we are dependent upon
            # have been resolved.
            for dep in package.ResolvedDirectDependencies:
                if dep.Access != AccessType.Link:
                    for dir in dep.Package.ResolvedBuildPublicIncludeDirs:
                        if dir == 'include':
                            dir = self.__ExtractIncludePath(config, dep.Package)
                        if not dir in includeDirs:
                            includeDirs.add(dir)
                            if dep.Access == AccessType.Private:
                                privateIncludeDirs.append(dir)
                            else:
                                publicIncludeDirs.append(dir)

            allIncludeDirs = list(includeDirs)
            allIncludeDirs.sort(key=lambda s: s.lower())
            publicIncludeDirs.sort(key=lambda s: s.lower())
            privateIncludeDirs.sort(key=lambda s: s.lower())
            if hasLocalIncludeDir:
                allIncludeDirs.insert(0, "include")
                publicIncludeDirs.insert(0, "include")
            package.ResolvedBuildPublicIncludeDirs = publicIncludeDirs
            package.ResolvedBuildPrivateIncludeDirs = privateIncludeDirs
            package.ResolvedBuildAllIncludeDirs = allIncludeDirs


    def __ResolveBuildCPPDefines(self, config, platformName, finalResolveOrder):
        # We proces the packages in the correct resolve order
        # This ensure that all dependencies have been processed 
        for package in finalResolveOrder:
            allCPPDefines = {}
            publicCPPDefines = []
            privateCPPDefines = []

            # Find the direct cpp defines in this package
            depCPPDefine = package.GetDirectCPPDefines(platformName)
            for cppDefine in depCPPDefine:
                processedCPPDefine = PackageCPPDefine(cppDefine, package.Name)
                if not package.IsVirtual:
                    processedCPPDefine.ConsumedBy = package
                    processedCPPDefine.IsFirstActualUse = True
                allCPPDefines[processedCPPDefine.Name] = processedCPPDefine
                if cppDefine.Access == AccessType.Public:
                    publicCPPDefines.append(processedCPPDefine)
                elif cppDefine.Access == AccessType.Private: # and depPackage.Name == package.Name:
                    privateCPPDefines.append(processedCPPDefine)
            
            package.ResolvedBuildDirectCPPDefines = allCPPDefines.values()

            directCPPDefines = dict(allCPPDefines)
            # Add the cpp defines from all direct dependencies
            for depPackage in package.ResolvedDirectDependencies:
                if depPackage.Access != AccessType.Link:
                    for entry in depPackage.Package.ResolvedBuildAllPublicCPPDefines:
                        processedCPPDefine = PackageCPPDefine(entry, entry.IntroducedByPackageName)
                        if not allCPPDefines.has_key(processedCPPDefine.Name):
                            allCPPDefines[processedCPPDefine.Name] = processedCPPDefine
                            if depPackage.Access == AccessType.Private:
                                privateCPPDefines.append(processedCPPDefine)
                            else:
                                publicCPPDefines.append(processedCPPDefine)
                            # If the cpp define has'nt been marked as consumed yet and this package isn't virtual
                            # then consume it
                            if processedCPPDefine.ConsumedBy == None and not package.IsVirtual:
                                processedCPPDefine.ConsumedBy = package
                                processedCPPDefine.IsFirstActualUse = True
                        elif directCPPDefines.has_key(processedCPPDefine.Name):
                            raise UsageErrorException("CPPDefine: %s was already defined by %s" % (processedCPPDefine.Name, processedCPPDefine.IntroducedByPackageName));

            allCPPDefines = list(allCPPDefines.values())
            allCPPDefines.sort(key=lambda s: s.Name.lower())
            publicCPPDefines.sort(key=lambda s: s.Name.lower())
            privateCPPDefines.sort(key=lambda s: s.Name.lower())
            package.ResolvedBuildAllCPPDefines = allCPPDefines
            package.ResolvedBuildAllPublicCPPDefines = publicCPPDefines
            package.ResolvedBuildAllPrivateCPPDefines = privateCPPDefines


    def __ResolveBuildUsedFeatures(self, config, platformName, finalResolveOrder):
        for package in finalResolveOrder:
            uniqueDict = {}
            for feature in package.GetDirectUsedFeatures(platformName):
                lowerName = feature.Name.lower()
                if uniqueDict.has_key(lowerName):
                    raise FeatureUseDuplicatedException(package, feature.Name)
                package.ResolvedDirectUsedFeatures.append(feature)
                uniqueDict[lowerName] = feature.Name

            uniqueFeatureNames = set()
            lowercaseNameDict = {}
            for depPackage in package.ResolvedDirectDependencies:
                for feature in depPackage.Package.ResolvedAllUsedFeatures:
                    if not feature.Name in uniqueFeatureNames:
                        lowerName = feature.Name.lower()
                        if lowercaseNameDict.has_key(lowerName):
                            raise FeatureNameCollisionException(depPackage.Package, feature.Name, lowercaseNameDict[lowerName][0], lowercaseNameDict[lowerName][1])
                        lowercaseNameDict[lowerName] = [depPackage.Package, feature.Name]
                        package.ResolvedAllUsedFeatures.append(feature)
                        uniqueFeatureNames.add(feature.Name)

            for feature in package.ResolvedDirectUsedFeatures:
                if not feature.Name in uniqueFeatureNames:
                    lowerName = feature.Name.lower()
                    if lowercaseNameDict.has_key(lowerName):
                        raise FeatureNameCollisionException(package, feature.Name, lowercaseNameDict[lowerName][0], lowercaseNameDict[lowerName][1])
                    lowercaseNameDict[lowerName] = [package, feature.Name]
                    package.ResolvedAllUsedFeatures.append(feature)
                    uniqueFeatureNames.add(feature.Name)

            package.ResolvedAllUsedFeatures.sort(key=lambda s: s.Name.lower())

    def __ExtractIncludePath(self, config, package):
        path = config.ToPath(package.AbsoluteIncludePath)
        return Util.UTF8ToAscii(path)


    def __ResolveDirectExternalDependencies(self, platformName, packages):
        for package in packages:
            package.ResolvedDirectExternalDependencies = package.GetExternalDependencies(platformName)
            #print("Deps: " + ", ".join(Util.ExtractNames(package.ResolvedDirectExternalDependencies)))


    def __ResolveDirectVariants(self, finalResolveOrder, platformName):
        # First we resolve all direct variants introduced by a package
        for package in finalResolveOrder:
            package.ResolvedDirectVariants = package.GetVariants(platformName)

    
    def __GetVariantUndefinedOptions(self, baseVariant, extendingVariant):
        res = []
        for entry in extendingVariant.Options:
            if not baseVariant.OptionDict.has_key(entry.Name):
                res.append(entry)
        return res

    # This is a fast way to validate various variant rules
    def __CheckVariants(self, config, finalResolveOrder):
        variantDict = {}
        for package in finalResolveOrder:
            # collect the variants the we encounter and check if any 
            # later encounter of it tries to extend it with new options 
            for variant in package.ResolvedDirectVariants:
                if not variantDict.has_key(variant.Name):
                    variantDict[variant.Name] = [package, variant];
                else:
                    record = variantDict[variant.Name]
                    if not config.AllowVariantExtension:
                        raise VariantExtensionNotSupportedException(package, variant, record[0])
                    undefinedOptions = self.__GetVariantUndefinedOptions(record[1], variant)
                    if len(undefinedOptions) > 0:
                        raise ExtendingVariantCanNotIntroduceNewOptionsException(package, variant, record[0], record[1], undefinedOptions)
        
        # Ensure that the variant names are unique when casing is ignored
        uniqueNames = {}
        for key, value in variantDict.iteritems():
            variantName = key.lower()
            if not uniqueNames.has_key(variantName):
                uniqueNames[variantName] = value
            else:
                record = uniqueNames[variantName]
                raise VariantNameCollisionException(value[0], value[1], record[0], record[1])


    def __ResolveAllVariants(self, finalResolveOrder):
        # For each package resolve all variants that it depends upon
        # NOTE: this resolver can be optimized by utilizing the fact that we resolve things in the correct order
        #       so all dependencies will have been resolved before the current package
        for package in finalResolveOrder:
            variantDict = {}
            for depPackage in package.ResolvedBuildOrder:
                for variant in depPackage.ResolvedDirectVariants:
                    clonedVariant = PackagePlatformVariant(variant, depPackage == package)
                    if not variantDict.has_key(variant.Name):
                        variantDict[variant.Name] = clonedVariant
                    else:
                        variantDict[variant.Name] = variantDict[variant.Name].Extend(clonedVariant, depPackage.Name)

            package.ResolvedAllVariantDict = variantDict

        # Update the directly resolved variants so they match the ones we utilize in the all list
        # we do this after everything has been resolved to prevent problems with modifying the ResolvedDirectVariants
        # that the above algorithm depends on
        for package in finalResolveOrder:
            variantDict = {}
            for variant in package.ResolvedAllVariantDict.values():
                variantDict[variant.Name] = variant
            for i in range(len(package.ResolvedDirectVariants)):
                variantName = package.ResolvedDirectVariants[i].Name
                if variantDict.has_key(variantName):
                    package.ResolvedDirectVariants[i] = variantDict[variantName]


    def __MarkExternalLibFirstUse(self, topLevelPackage):
#        self.__MarkExternalLibFirstUse2(topLevelPackage)
        for variant in topLevelPackage.ResolvedAllVariantDict.values():
            for option in variant.Options:
                for dep in topLevelPackage.ResolvedDirectDependencies:
                    self.__MarkExternalLibFirstUse2(dep.Package, variant.Name, option.Name)


    def __MarkExternalLibFirstUse2(self, package, variantName, optionName):
        if package.IsVirtual or not variantName in package.ResolvedAllVariantDict.keys():
            return set()
        
        usedSet = set()
        for dep in package.ResolvedDirectDependencies:
            usedSet |= self.__MarkExternalLibFirstUse2(dep.Package, variantName, optionName)

        option = package.ResolvedAllVariantDict[variantName].OptionDict[optionName]
        for entry in option.ExternalDependencies:
            if not entry.Name in usedSet:
                usedSet.add(entry.Name)
                entry.IsFirstActualUse = True
        return usedSet


    def __ResolvePlatform(self, platformName, packages):
        for package in packages:
            package.ResolvedPlatformName = platformName
            package.ResolvedPlatform = package.GetPlatform(platformName)
            if package.ResolvedPlatform != None:
                # use the flag set in the xml
                package.ResolvedPlatformNotSupported = package.ResolvedPlatform.NotSupported
            else:
                package.ResolvedPlatformNotSupported = False
        
        if platformName == PackageConfig.PLATFORM_WINDOWS:
            for package in packages:
                if package.Type == PackageType.Library or package.Type == PackageType.Executable: # or package.Type == PackageType.ExeLibCombo:
                    if package.ResolvedPlatform != None and package.ResolvedPlatform.ProjectId == None:
                        raise XmlMissingWindowsVisualStudioProjectIdException(package.XMLElement)
