#!/usr/bin/env python

#****************************************************************************************************************************************************
# Copyright (c) 2014 Freescale Semiconductor, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#
#    * Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#
#    * Neither the name of the Freescale Semiconductor, Inc. nor the names of
#      its contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#****************************************************************************************************************************************************

import copy
import os
import os.path
import xml.etree.ElementTree as ET
from FslBuildGen.DataTypes import *
from FslBuildGen.Exceptions import *
from FslBuildGen.XmlStuff import *
from FslBuildGen import IOUtil, Util, PackageConfig
from FslBuildGen.XmlToolConfigFile import XmlToolConfigFile


class ToolConfigTemplateFolder(object):
    def __init__(self, basedUponXML):
        super(ToolConfigTemplateFolder, self).__init__()
        self.BasedOn = basedUponXML
        self.Name = self.BasedOn.Name
        self.ResolvedPath = IOUtil.ToUnixStylePath(self.__ResolvePath(self.Name))


    def __ExtractEnvironmentVariablesFromName(self, entry):
        if entry.startswith("$("):
            endIndex = entry.find(")")
            if endIndex < 0:
                raise XmlException2(entry.XmlElement, "Incomplete environment variable found");
            return entry[2:endIndex]
        elif entry.find("$(") >= 0:
            raise XmlException2(entry.XmlElement, "Environment variables can not be located in the middle of a string");
        return None

    def __ResolvePath(self, entry):
        name = self.Name
        environmentName = self.__ExtractEnvironmentVariablesFromName(name)
        if environmentName != None:
            path = IOUtil.GetEnvironmentVariableForDirectory(environmentName)
            name = name.replace("$("+environmentName+")", path)
        elif not os.path.isabs(name):
            raise EnvironmentError("%s template folder path is not absolute" % (name))

        if not os.path.isdir(name):
            raise EnvironmentError("%s template folder is not a valid directory" % name)
        return name


class ToolConfigRootDirectory(object):
    def __init__(self, basedUponXML):
        super(ToolConfigRootDirectory, self).__init__()
        self.BasedOn = basedUponXML
        self.Name = self.BasedOn.Name
        self.DecodedName = self.__ExtractEnvironmentVariablesFromName(self.Name)
        if self.DecodedName == None:
            raise XmlException2(basedUponXML.XmlElement, "Root dirs are expected to contain environment variables");
        self.BashName = '$' + self.DecodedName
        self.DosName = '%' + self.DecodedName + '%'
        self.ResolvedPath = IOUtil.ToUnixStylePath(IOUtil.GetEnvironmentVariableForDirectory(self.DecodedName))
        self.ResolvedPathEx = self.ResolvedPath + '/'


    def __ExtractEnvironmentVariablesFromName(self, entry):
        if entry.startswith("$("):
            endIndex = entry.find(")")
            if endIndex < 0:
                raise XmlException2(entry.XmlElement, "Incomplete environment variable found");
            elif endIndex != (len(entry)-1):
                raise XmlException2(entry.XmlElement, "The environment variable can not be combined with another path");
            return entry[2:len(entry)-1]
        elif entry.find("$(") >= 0:
            raise XmlException2(entry.XmlElement, "Environment variables can not be located in the middle of a string");
        return None


class ToolConfigDirectory(object):
    def __init__(self, basedUponXML):
        super(ToolConfigDirectory, self).__init__()

        self.BasedOn = basedUponXML
        self.Name = self.BasedOn.Name

        envName, rest  = self.__SplitEnvironmentVariablesFromName(self.Name)
        if envName == None:
            raise XmlException2(basedUponXML.XmlElement, "Template import dirs are expected to contain environment variables");

        self.DecodedName = envName
        self.BashName = IOUtil.Join('$' + self.DecodedName, rest)
        self.DosName = IOUtil.Join('%' + self.DecodedName + '%', rest)
        if self.Name == None:
            raise XmlException2(basedUponXML.XmlElement, "Dirs are expected to contain environment variables");
        self.ResolvedPath = IOUtil.Join(IOUtil.GetEnvironmentVariableForDirectory(self.DecodedName), rest)
        self.ResolvedPathEx = self.ResolvedPath + '/'


    def __SplitEnvironmentVariablesFromName(self, entry):
        if entry.startswith("$("):
            endIndex = entry.find(")")
            if endIndex < 0:
                raise XmlException2(entry.XmlElement, "Incomplete environment variable found");

            env = entry[2:endIndex]
            rest = entry[endIndex+1:] if (endIndex+1) <= len(entry) else ''
            if rest.startswith('/') or rest.startswith('\\'):
                rest = rest[1:]
            return env, rest
        elif entry.find("$(") >= 0:
            raise XmlException2(entry.XmlElement, "Environment variables can not be located in the middle of a string");
        return None, None


class ToolConfigPackageLocation(object):
    def __init__(self, rootDirs, basedUponXML):
        super(ToolConfigPackageLocation, self).__init__()
        self.BasedOn = basedUponXML
        self.Name = basedUponXML.Name
        self.Blacklist = basedUponXML.Blacklist
        self.ResolvedPath = self.__ResolvePath(rootDirs, self)

    def __ResolvePath(self, rootDirs, entry):
        rootDir = self.__LocateRootDir(rootDirs, entry)
        return entry.Name.replace(rootDir.Name, rootDir.ResolvedPath)

    def __LocateRootDir(self, rootDirs, entry):
        for rootDir in rootDirs:
            if entry.Name.startswith(rootDir.Name):
                return rootDir
        raise XmlException2(entry.BasedOn.XMLElement, "Path '%s' is not based on one of the valid root directories %s" % (entry.Name, ", ".join(Util.ExtractNames(rootDirs))))



class ToolConfigPackageConfiguration(object):
    def __init__(self, rootDirs, basedUponXML):
        super(ToolConfigPackageConfiguration, self).__init__()
        self.BasedOn = basedUponXML
        self.Name = basedUponXML.Name
        self.Locations = self.__ResolveLocations(rootDirs, basedUponXML.Locations)

    def __ResolveLocations(self, rootDirs, locations):
        res = []
        for location in locations:
            res.append(ToolConfigPackageLocation(rootDirs, location))
        return res


class ToolConfig(object):
    def __init__(self, filename):
        super(ToolConfig, self).__init__()
        basedUponXML = XmlToolConfigFile(filename)
        self.BasedOn = basedUponXML
        self.GenFileName = basedUponXML.GenFileName.Name
        self.RootDirectories = self.__ResolveRootDirectories(basedUponXML.RootDirectories)
        self.TemplateImportDirectories = self.__ResolveDirectories(basedUponXML.TemplateImportDirectories)
        self.PackageConfiguration = self.__ResolvePackageConfiguration(self.RootDirectories, basedUponXML.PackageConfiguration)
        self.TemplateFolder = ToolConfigTemplateFolder(basedUponXML.TemplateFolder)
        self.UnitTestPath = self.__ResolveUnitTestPath()


    def ToPath(self, path):
        if path == None:
            return None
        if path.find("\\") >= 0:
            raise UsageErrorException("Backslash found in the supplied path '%s'" % (path))
        for rootDir in self.RootDirectories:
            if path.startswith(rootDir.ResolvedPathEx):
                lenRootPath = len(rootDir.ResolvedPathEx)
                path = path[lenRootPath:]
                return rootDir.Name + "/" + Util.UTF8ToAscii(path)
        raise UsageErrorException("ERROR: the folder '%s' does not reside inside one of the root dirs" % (path))


    def ToBashPath(self, path):
        if path == None:
            return None
        if path.find("\\") >= 0:
            raise UsageErrorException("Backslash found in the supplied path '%s'" % (path))
        for rootDir in self.RootDirectories:
            if path.startswith(rootDir.ResolvedPathEx):
                lenRootPath = len(rootDir.ResolvedPathEx)
                path = path[lenRootPath:]
                return rootDir.BashName + "/" + Util.UTF8ToAscii(path)
        raise UsageErrorException("ERROR: the folder '%s' does not reside inside one of the root dirs" % (path))


    def ToDosPath(self, path):
        if path == None:
            return None
        if path.find("\\") >= 0:
            raise UsageErrorException("Backslash found in the supplied path '%s'" % (path))
        for rootDir in self.RootDirectories:
            if path.startswith(rootDir.ResolvedPathEx):
                lenRootPath = len(rootDir.ResolvedPathEx)
                path = path[lenRootPath:]
                tmp = rootDir.DosName + "/" + Util.UTF8ToAscii(path)
                return tmp.replace('/', '\\')
        raise UsageErrorException("ERROR: the folder '%s' does not reside inside one of the root dirs" % (path))


    def __ResolveRootDirectories(self, rootDirectories):
        rootDirs = []
        for rootDir in rootDirectories:
            rootDirs.append(ToolConfigRootDirectory(rootDir))
        return rootDirs

    def __ResolveDirectories(self, directories):
        dirs = []
        for dir in directories:
            dirs.append(ToolConfigDirectory(dir))
        return dirs


    def __ResolvePackageConfiguration(self, rootDirs, packageConfiguration):
        configs = {}
        for packageConfig in packageConfiguration.values():
            resolvedConfig = ToolConfigPackageConfiguration(rootDirs, packageConfig)
            configs[resolvedConfig.Name] = resolvedConfig
        return configs

    def __ResolveUnitTestPath(self):
        path = os.environ.get("FSL_GRAPHICS_INTERNAL")
        if path == None:
            return None
        return IOUtil.Join(path, "Tools/FslBuildGen/FslBuildGen/UnitTest/TestFiles")

