/****************************************************************************************************************************************************
* Copyright (c) 2014 Freescale Semiconductor, Inc.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*    * Redistributions of source code must retain the above copyright notice,
*      this list of conditions and the following disclaimer.
*
*    * Redistributions in binary form must reproduce the above copyright notice,
*      this list of conditions and the following disclaimer in the documentation
*      and/or other materials provided with the distribution.
*
*    * Neither the name of the Freescale Semiconductor, Inc. nor the names of
*      its contributors may be used to endorse or promote products derived from
*      this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
* IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
* BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
* OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
****************************************************************************************************************************************************/

#include <FslBase/Bits/ByteArrayUtil.hpp>
#include <cassert>

namespace Fsl
{
  uint16_t ByteArrayUtil::ReadUInt16LE(const uint8_t*const pSrc, const int32_t srcLength, const int32_t index)
  {
    assert(pSrc != nullptr);
    assert(srcLength >= 2);
    assert(index >= 0 && (index + 1) < srcLength);

    return static_cast<uint16_t>(static_cast<uint16_t>(pSrc[index]) | (static_cast<uint16_t>(pSrc[index + 1]) << 8));
  }


  int16_t ByteArrayUtil::ReadInt16LE(const uint8_t*const pSrc, const int32_t srcLength, const int32_t index)
  {
    assert(pSrc != nullptr);
    assert(srcLength >= 2);
    assert(index >= 0 && (index + 1) < srcLength);

    return static_cast<int16_t>(static_cast<uint16_t>(pSrc[index]) | (static_cast<uint16_t>(pSrc[index + 1]) << 8));
  }


  uint32_t ByteArrayUtil::ReadUInt32LE(const uint8_t*const pSrc, const int32_t srcLength, const int32_t index)
  {
    assert(pSrc != nullptr);
    assert(srcLength >= 4);
    assert(index >= 0 && (index + 3) < srcLength);

    return static_cast<uint32_t>(static_cast<uint32_t>(pSrc[index])            | (static_cast<uint32_t>(pSrc[index + 1]) << 8) |
                                (static_cast<uint32_t>(pSrc[index + 2]) << 16) | (static_cast<uint32_t>(pSrc[index + 3]) << 24));
  }


  int32_t ByteArrayUtil::ReadInt32LE(const uint8_t*const pSrc, const int32_t srcLength, const int32_t index)
  {
    assert(pSrc != nullptr);
    assert(srcLength >= 4);
    assert(index >= 0 && (index + 3) < srcLength);

    return static_cast<int32_t>(static_cast<uint32_t>(pSrc[index])            | (static_cast<uint32_t>(pSrc[index + 1]) << 8) |
                               (static_cast<uint32_t>(pSrc[index + 2]) << 16) | (static_cast<uint32_t>(pSrc[index + 3]) << 24));
  }


  //uint64_t ByteArrayUtil::ReadUInt64LE(const uint8_t*const pSrc, const int32_t srcLength, const int32_t index)
  //{

  //}

  //int64_t ByteArrayUtil::ReadInt64LE(const uint8_t*const pSrc, const int32_t srcLength, const int32_t index)
  //{

  //}


  uint16_t ByteArrayUtil::ReadUInt16BE(const uint8_t*const pSrc, const int32_t srcLength, const int32_t index)
  {
    assert(pSrc != nullptr);
    assert(srcLength >= 2);
    assert(index >= 0 && (index + 1) < srcLength);


    return static_cast<uint16_t>(static_cast<uint16_t>(pSrc[index + 1]) | (static_cast<uint16_t>(pSrc[index]) << 8));
  }


  int16_t ByteArrayUtil::ReadInt16BE(const uint8_t*const pSrc, const int32_t srcLength, const int32_t index)
  {
    assert(pSrc != nullptr);
    assert(srcLength >= 2);
    assert(index >= 0 && (index + 1) < srcLength);

    return static_cast<int16_t>(static_cast<uint16_t>(pSrc[index + 1]) | (static_cast<uint16_t>(pSrc[index]) << 8));
  }


  uint32_t ByteArrayUtil::ReadUInt32BE(const uint8_t*const pSrc, const int32_t srcLength, const int32_t index)
  {
    assert(pSrc != nullptr);
    assert(srcLength >= 4);
    assert(index >= 0 && (index + 3) < srcLength);


    return static_cast<uint32_t>(static_cast<uint32_t>(pSrc[index + 3])        | (static_cast<uint32_t>(pSrc[index + 2]) << 8) |
                                (static_cast<uint32_t>(pSrc[index + 1]) << 16) | (static_cast<uint32_t>(pSrc[index    ]) << 24));
  }


  int32_t ByteArrayUtil::ReadInt32BE(const uint8_t*const pSrc, const int32_t srcLength, const int32_t index)
  {
    assert(pSrc != nullptr);
    assert(srcLength >= 4);
    assert(index >= 0 && (index + 3) < srcLength);


    return static_cast<int32_t>(static_cast<uint32_t>(pSrc[index + 3])        | (static_cast<uint32_t>(pSrc[index + 2]) << 8) |
                               (static_cast<uint32_t>(pSrc[index + 1]) << 16) | (static_cast<uint32_t>(pSrc[index]) << 24));

  }


  //uint64_t ByteArrayUtil::ReadUInt64BE(const uint8_t*const pSrc, const int32_t srcLength, const int32_t index)
  //{

  //}


  //int64_t ByteArrayUtil::ReadInt64BE(const uint8_t*const pSrc, const int32_t srcLength, const int32_t index)
  //{

  //}


  void ByteArrayUtil::WriteUInt16LE(uint8_t* pDst, const int32_t dstLength, const int32_t index, const uint16_t value)
  {
    assert(pDst != nullptr);
    assert(dstLength >= 2);
    assert(index >= 0 && (index + 1) < dstLength);


    pDst[index] = static_cast<uint8_t>(value & 0xFF);
    pDst[index + 1] = static_cast<uint8_t>((value >> 8) & 0xFF);
  }


  void ByteArrayUtil::WriteInt16LE(uint8_t* pDst, const int32_t dstLength, const int32_t index, const int16_t value)
  {
    assert(pDst != nullptr);
    assert(dstLength >= 2);
    assert(index >= 0 && (index + 1) < dstLength);


    pDst[index] = static_cast<uint8_t>(value & 0xFF);
    pDst[index + 1] = static_cast<uint8_t>((value >> 8) & 0xFF);
  }


  void ByteArrayUtil::WriteUInt32LE(uint8_t* pDst, const int32_t dstLength, const int32_t index, const uint32_t value)
  {
    assert(pDst != nullptr);
    assert(dstLength >= 4);
    assert(index >= 0 && (index + 3) < dstLength);


    pDst[index] = static_cast<uint8_t>(value & 0xFF);
    pDst[index + 1] = static_cast<uint8_t>((value >> 8) & 0xFF);
    pDst[index + 2] = static_cast<uint8_t>((value >> 16) & 0xFF);
    pDst[index + 3] = static_cast<uint8_t>((value >> 24) & 0xFF);
  }


  void ByteArrayUtil::WriteInt32LE(uint8_t* pDst, const int32_t dstLength, const int32_t index, const int32_t value)
  {
    assert(pDst != nullptr);
    assert(dstLength >= 4);
    assert(index >= 0 && (index + 3) < dstLength);


    pDst[index] = static_cast<uint8_t>(value & 0xFF);
    pDst[index + 1] = static_cast<uint8_t>((value >> 8) & 0xFF);
    pDst[index + 2] = static_cast<uint8_t>((value >> 16) & 0xFF);
    pDst[index + 3] = static_cast<uint8_t>((value >> 24) & 0xFF);
  }


  //void ByteArrayUtil::WriteUInt64LE(uint8_t* pDst, const int32_t dstLength, const int32_t index, const uint64_t value)
  //{

  //}


  //void ByteArrayUtil::WriteInt64LE(uint8_t* pDst, const int32_t dstLength, const int32_t index, const int64_t value)
  //{

  //}


  void ByteArrayUtil::WriteUInt16BE(uint8_t* pDst, const int32_t dstLength, const int32_t index, const uint16_t value)
  {
    assert(pDst != nullptr);
    assert(dstLength >= 2);
    assert(index >= 0 && (index + 1) < dstLength);


    pDst[index + 0] = static_cast<uint8_t>((value >> 8) & 0xFF);
    pDst[index + 1] = static_cast<uint8_t>((value)& 0xFF);
  }


  void ByteArrayUtil::WriteInt16BE(uint8_t* pDst, const int32_t dstLength, const int32_t index, const int16_t value)
  {
    assert(pDst != nullptr);
    assert(dstLength >= 2);
    assert(index >= 0 && (index + 1) < dstLength);


    pDst[index + 0] = static_cast<uint8_t>((value >> 8) & 0xFF);
    pDst[index + 1] = static_cast<uint8_t>((value)& 0xFF);
  }


  void ByteArrayUtil::WriteUInt32BE(uint8_t* pDst, const int32_t dstLength, const int32_t index, const uint32_t value)
  {
    assert(pDst != nullptr);
    assert(dstLength >= 4);
    assert(index >= 0 && (index + 3) < dstLength);


    pDst[index + 0] = static_cast<uint8_t>((value >> 24) & 0xFF);
    pDst[index + 1] = static_cast<uint8_t>((value >> 16) & 0xFF);
    pDst[index + 2] = static_cast<uint8_t>((value >> 8) & 0xFF);
    pDst[index + 3] = static_cast<uint8_t>((value)& 0xFF);
  }


  void ByteArrayUtil::WriteInt32BE(uint8_t* pDst, const int32_t dstLength, const int32_t index, const int32_t value)
  {
    assert(pDst != nullptr);
    assert(dstLength >= 4);
    assert(index >= 0 && (index + 3) < dstLength);

    pDst[index + 0] = static_cast<uint8_t>((value >> 24) & 0xFF);
    pDst[index + 1] = static_cast<uint8_t>((value >> 16) & 0xFF);
    pDst[index + 2] = static_cast<uint8_t>((value >> 8) & 0xFF);
    pDst[index + 3] = static_cast<uint8_t>((value)& 0xFF);
  }


  //void ByteArrayUtil::WriteUInt64BE(uint8_t* pDst, const int32_t dstLength, const int32_t index, const uint64_t value)
  //{

  //}


  //void ByteArrayUtil::WriteInt64BE(uint8_t* pDst, const int32_t dstLength, const int32_t index, const int32_t value)
  //{

  //}


  void ByteArrayUtil::WriteLE(uint8_t* pDst, const int32_t dstLength, const int32_t index, const uint16_t value)
  {
    assert(pDst != nullptr);
    assert(dstLength >= 2);
    assert(index >= 0 && (index + 1) < dstLength);

    pDst[index] = static_cast<uint8_t>(value & 0xFF);
    pDst[index + 1] = static_cast<uint8_t>((value >> 8) & 0xFF);
  }


  void ByteArrayUtil::WriteLE(uint8_t* pDst, const int32_t dstLength, const int32_t index, const int16_t value)
  {
    assert(pDst != nullptr);
    assert(dstLength >= 2);
    assert(index >= 0 && (index + 1) < dstLength);

    pDst[index] = static_cast<uint8_t>(value & 0xFF);
    pDst[index + 1] = static_cast<uint8_t>((value >> 8) & 0xFF);
  }


  void ByteArrayUtil::WriteLE(uint8_t* pDst, const int32_t dstLength, const int32_t index, const uint32_t value)
  {
    assert(pDst != nullptr);
    assert(dstLength >= 4);
    assert(index >= 0 && (index + 3) < dstLength);

    pDst[index] = static_cast<uint8_t>(value & 0xFF);
    pDst[index + 1] = static_cast<uint8_t>((value >> 8) & 0xFF);
    pDst[index + 2] = static_cast<uint8_t>((value >> 16) & 0xFF);
    pDst[index + 3] = static_cast<uint8_t>((value >> 24) & 0xFF);
  }


  void ByteArrayUtil::WriteLE(uint8_t* pDst, const int32_t dstLength, const int32_t index, const int32_t value)
  {
    assert(pDst != nullptr);
    assert(dstLength >= 4);
    assert(index >= 0 && (index + 3) < dstLength);

    pDst[index] = static_cast<uint8_t>(value & 0xFF);
    pDst[index + 1] = static_cast<uint8_t>((value >> 8) & 0xFF);
    pDst[index + 2] = static_cast<uint8_t>((value >> 16) & 0xFF);
    pDst[index + 3] = static_cast<uint8_t>((value >> 24) & 0xFF);
  }


  //void ByteArrayUtil::WriteLE(uint8_t* pDst, const int32_t dstLength, const int32_t index, const uint64_t value)
  //{

  //}


  //void ByteArrayUtil::WriteLE(uint8_t* pDst, const int32_t dstLength, const int32_t index, const int64_t value)
  //{

  //}


  void ByteArrayUtil::WriteBE(uint8_t* pDst, const int32_t dstLength, const int32_t index, const uint16_t value)
  {
    assert(pDst != nullptr);
    assert(dstLength >= 2);
    assert(index >= 0 && (index + 1) < dstLength);

    pDst[index + 0] = static_cast<uint8_t>((value >> 8) & 0xFF);
    pDst[index + 1] = static_cast<uint8_t>((value) & 0xFF);
  }


  void ByteArrayUtil::WriteBE(uint8_t* pDst, const int32_t dstLength, const int32_t index, const int16_t value)
  {
    assert(pDst != nullptr);
    assert(dstLength >= 2);
    assert(index >= 0 && (index + 1) < dstLength);

    pDst[index + 0] = static_cast<uint8_t>((value >> 8) & 0xFF);
    pDst[index + 1] = static_cast<uint8_t>((value) & 0xFF);
  }


  void ByteArrayUtil::WriteBE(uint8_t* pDst, const int32_t dstLength, const int32_t index, const uint32_t value)
  {
    assert(pDst != nullptr);
    assert(dstLength >= 4);
    assert(index >= 0 && (index + 3) < dstLength);

    pDst[index + 0] = static_cast<uint8_t>((value >> 24) & 0xFF);
    pDst[index + 1] = static_cast<uint8_t>((value >> 16) & 0xFF);
    pDst[index + 2] = static_cast<uint8_t>((value >> 8) & 0xFF);
    pDst[index + 3] = static_cast<uint8_t>((value) & 0xFF);
  }


  void ByteArrayUtil::WriteBE(uint8_t* pDst, const int32_t dstLength, const int32_t index, const int32_t value)
  {
    assert(pDst != nullptr);
    assert(dstLength >= 4);
    assert(index >= 0 && (index + 3) < dstLength);

    pDst[index + 0] = static_cast<uint8_t>((value >> 24) & 0xFF);
    pDst[index + 1] = static_cast<uint8_t>((value >> 16) & 0xFF);
    pDst[index + 2] = static_cast<uint8_t>((value >> 8) & 0xFF);
    pDst[index + 3] = static_cast<uint8_t>((value)& 0xFF);
  }


  //void ByteArrayUtil::WriteBE(uint8_t* pDst, const int32_t dstLength, const int32_t index, const uint64_t value)
  //{

  //}


  //void ByteArrayUtil::WriteBE(uint8_t* pDst, const int32_t dstLength, const int32_t index, const int32_t value)
  //{

  //}


}
