/****************************************************************************************************************************************************
* Copyright (c) 2014 Freescale Semiconductor, Inc.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*    * Redistributions of source code must retain the above copyright notice,
*      this list of conditions and the following disclaimer.
*
*    * Redistributions in binary form must reproduce the above copyright notice,
*      this list of conditions and the following disclaimer in the documentation
*      and/or other materials provided with the distribution.
*
*    * Neither the name of the Freescale Semiconductor, Inc. nor the names of
*      its contributors may be used to endorse or promote products derived from
*      this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
* IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
* BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
* OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
****************************************************************************************************************************************************/

#include <FslDemoApp/ADemoApp.hpp>
#include <FslDemoApp/Service/Exceptions.hpp>
#include <FslDemoApp/Service/Events/IEvent.hpp>
#include <FslDemoApp/Service/Events/Basic/KeyEvent.hpp>
#include <cassert>

namespace Fsl
{
  namespace
  {
    void ToggleTimestep(const std::shared_ptr<IDemoAppControl> demoAppControl, const TimeStepMode::Enum mode)
    {
      const TimeStepMode::Enum newTimeStepMode = demoAppControl->GetTimeStepMode() != mode ? mode : TimeStepMode::Normal;
      demoAppControl->SetTimeStepMode(newTimeStepMode);
    }

  }


  ADemoApp::ADemoApp(const DemoAppConfig& demoAppConfig)
    : m_demoAppConfig(demoAppConfig)
    , m_contentManger()
    , m_demoAppControl()
  {
    m_contentManger = demoAppConfig.DemoServiceProvider.Get<IContentManager>();
    m_demoAppControl = demoAppConfig.DemoServiceProvider.Get<IDemoAppControl>();
  }


  void ADemoApp::_OnEvent(IEvent*const pEvent)
  {
    // Done this way to prevent common mistakes where people forget to call the base class
    const EventType::Enum eventType = pEvent->GetEventType();

    if ((eventType & EventType::ComplexEvent) == 0)
    {
      BasicEvent* pBasicEvent = dynamic_cast<BasicEvent*>(pEvent);
      assert(pBasicEvent != nullptr);

      switch (pEvent->GetEventType())
      {
      case EventType::KeyPressed:
      {
        KeyEvent keyEvent(*pBasicEvent);
        OnKeyEvent(keyEvent);
        if (!keyEvent.IsHandled())
          UnhandledKeyFallback(keyEvent);
        break;
      }
      case EventType::MouseButton:
        OnMouseButtonEvent(MouseButtonEvent(*pBasicEvent));
        break;
      case EventType::MouseMove:
        OnMouseMoveEvent(MouseMoveEvent(*pBasicEvent));
        break;
      case EventType::MouseWheel:
        OnMouseWheelEvent(MouseWheelEvent(*pBasicEvent));
        break;
      default:
        break;
      }
    }

  }


  void ADemoApp::_Resized(const Point2& size)
  {
    m_demoAppConfig.ScreenResolution = size;
    // Done this way to prevent common mistakes where people forget to call the base class
    Resized(size);
  }


  void ADemoApp::_FixedUpdate(const DemoTime& demoTime)
  {
    // Done this way to prevent common mistakes where people forget to call the base class
    FixedUpdate(demoTime);
  }


  void ADemoApp::_Update(const DemoTime& demoTime)
  {
    // Done this way to prevent common mistakes where people forget to call the base class
    Update(demoTime);
  }


  void ADemoApp::_Draw()
  {
    // Done this way to prevent common mistakes where people forget to call the base class
    Draw();
  }


  Point2 ADemoApp::GetScreenResolution() const
  {
    return m_demoAppConfig.ScreenResolution;
  }


  std::shared_ptr<IDemoAppControl> ADemoApp::GetDemoAppControl() const
  {
    const std::shared_ptr<IDemoAppControl> demoAppControl = m_demoAppControl.lock();
    if (!demoAppControl)
      throw ServiceUnavailableException("The service is no longer available");
    return demoAppControl;
  }


  std::shared_ptr<IContentManager> ADemoApp::GetContentManager() const
  {
    const std::shared_ptr<IContentManager> contentManager = m_contentManger.lock();
    if (!contentManager)
      throw ServiceUnavailableException("The service is no longer available");
    return contentManager;
  }


  void ADemoApp::UnhandledKeyFallback(const KeyEvent& keyEvent)
  {
    if (!keyEvent.IsPressed())
      return;


    const std::shared_ptr<IDemoAppControl> demoAppControl = GetDemoAppControl();
    switch (keyEvent.GetKey())
    {
    case VirtualKey::Escape:
      demoAppControl->RequestExit();
      break;
    case VirtualKey::Pause:
      ToggleTimestep(demoAppControl, TimeStepMode::Paused);
      break;
    case VirtualKey::PageDown:
      demoAppControl->SetTimeStepMode(TimeStepMode::Step);
      break;
    case VirtualKey::Delete:
      ToggleTimestep(demoAppControl, TimeStepMode::Slow2X);
      break;
    case VirtualKey::End:
      ToggleTimestep(demoAppControl, TimeStepMode::Slow4X);
      break;
    case VirtualKey::Insert:
      ToggleTimestep(demoAppControl, TimeStepMode::Fast2X);
      break;
    case VirtualKey::Home:
      ToggleTimestep(demoAppControl, TimeStepMode::Fast4X);
      break;
    default:
      break;
    }
  }

}

