/****************************************************************************************************************************************************
* Copyright (c) 2014 Freescale Semiconductor, Inc.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*    * Redistributions of source code must retain the above copyright notice,
*      this list of conditions and the following disclaimer.
*
*    * Redistributions in binary form must reproduce the above copyright notice,
*      this list of conditions and the following disclaimer in the documentation
*      and/or other materials provided with the distribution.
*
*    * Neither the name of the Freescale Semiconductor, Inc. nor the names of
*      its contributors may be used to endorse or promote products derived from
*      this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
* IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
* BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
* OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
****************************************************************************************************************************************************/

#include <FslGraphics/Exceptions.hpp>
#include <FslGraphics/Vertices/VertexDeclaration.hpp>
#include <FslGraphics/Vertices/VertexElementFormatUtil.hpp>
#include <algorithm>
#include <cassert>

namespace Fsl
{
  namespace
  {
    bool Contains(const std::vector<VertexElementEx>& elements, const VertexElementUsage::Enum usage, const int usageIndex, const int ignoreIndex)
    {
      const int count = int(elements.size());
      assert(count >= 0);
      for (int i = 0; i < count; ++i)
      {
        if (i != ignoreIndex && elements[i].Usage == usage && elements[i].UsageIndex == usageIndex)
          return true;
      }
      return false;
    }

    void VerifyElements(const std::vector<VertexElementEx>& elements, const int vertexStride)
    {
      // We expect the element offsets to be in order: smallest -> largest
      int maxOffset = 0;
      const int count = int(elements.size());
      assert(count >= 0);
      for (int i = 0; i < count; ++i)
      {
        int offset = elements[i].Offset + VertexElementFormatUtil::GetBytesPerElement(elements[i].Format);
        if (offset < maxOffset)
          throw NotSupportedException("The offsets should be in order: smallest -> largest");

        maxOffset = offset;

        if (Contains(elements, elements[i].Usage, elements[i].UsageIndex, i))
          throw NotSupportedException("Duplicated elements: A usage and usageIndex can only be listed once");
      }
      // FIX: implement a proper validation check of the elements here

      if ( maxOffset > vertexStride )
        throw NotSupportedException("At least one of the elements is outside of the vertex stride");
      if (vertexStride < 0 )
        throw NotSupportedException("Invalid vertex stride");
    }

    void CopyElements(std::vector<VertexElementEx>& rDstElements, const VertexElementEx* pSrcElement, const int elementCount)
    {
      assert(elementCount >= 0);
      assert(rDstElements.size() == size_t(elementCount));

      int maxOffset = 0;
      int dstIndex = 0;
      for (int i = 0; i < elementCount; ++i)
      {
        const int currentOffset = pSrcElement[i].Offset;
        if (currentOffset >= maxOffset)
        {
          maxOffset = currentOffset;
          rDstElements[dstIndex] = pSrcElement[i];
        }
        else
        {
          // back track and make room until we find a offset that is smaller
          int tmpIndex = dstIndex - 1;
          while (tmpIndex >= 0 && currentOffset < rDstElements[tmpIndex].Offset)
          {
            rDstElements[tmpIndex + 1] = rDstElements[tmpIndex];
            --tmpIndex;
          }
          if (tmpIndex < 0)
            tmpIndex = 0;
          rDstElements[tmpIndex] = pSrcElement[i];
        }
        ++dstIndex;
      }
    }
  }


  VertexDeclaration::VertexDeclaration()
    : m_elements()
    , m_vertexStride(0)
  {
  }


  VertexDeclaration::VertexDeclaration(const VertexElement*const pElements, const int elementCount)
    : m_elements(std::max(elementCount,0))
    , m_vertexStride(0)
  {
    if (pElements == nullptr)
      throw std::invalid_argument("pElements can not be null");
    if (elementCount < 0)
      throw std::invalid_argument("Element count can not be negative");

    int offset = 0;
    for (int i = 0; i < elementCount; ++i)
    {
      m_elements[i] = VertexElementEx(offset, pElements[i].Format, pElements[i].Usage, pElements[i].UsageIndex);
      offset += VertexElementFormatUtil::GetBytesPerElement(pElements[i].Format);
    }
    m_vertexStride = offset;

    VerifyElements(m_elements, m_vertexStride);
  }


  VertexDeclaration::VertexDeclaration(const VertexElementEx*const pElements, const int elementCount, const int vertexStride)
    : m_elements(std::max(elementCount, 0))
    , m_vertexStride(vertexStride)
  {
    if (pElements == nullptr)
      throw std::invalid_argument("pElements can not be null");
    if (elementCount < 0)
      throw std::invalid_argument("Element count can not be negative");

    CopyElements(m_elements, pElements, elementCount);
    VerifyElements(m_elements, m_vertexStride);
  }


  void VertexDeclaration::Reset(const VertexElement*const pElements, const int elementCount)
  {
    if (pElements == nullptr)
      throw std::invalid_argument("pElements can not be null");
    if (elementCount < 0)
      throw std::invalid_argument("Element count can not be negative");

    m_elements.resize(elementCount);

    int offset = 0;
    for (int i = 0; i < elementCount; ++i)
    {
      m_elements[i] = VertexElementEx(offset, pElements[i].Format, pElements[i].Usage, pElements[i].UsageIndex);
      offset += VertexElementFormatUtil::GetBytesPerElement(pElements[i].Format);
    }
    m_vertexStride = offset;

    VerifyElements(m_elements, m_vertexStride);
  }


  void VertexDeclaration::Reset(const VertexElementEx*const pElements, const int elementCount, const int vertexStride)
  {
    if (pElements == nullptr)
      throw std::invalid_argument("pElements can not be null");
    if (elementCount < 0)
      throw std::invalid_argument("Element count can not be negative");

    m_elements.resize(elementCount);

    CopyElements(m_elements, pElements, elementCount);
    VerifyElements(m_elements, m_vertexStride);
  }


  int VertexDeclaration::Count() const
  {
    return m_elements.size();
  }


  const VertexElementEx* VertexDeclaration::DirectAccess() const
  {
    return m_elements.data();
  }


  int VertexDeclaration::VertexElementIndexOf(const VertexElementUsage::Enum usage, const int usageIndex) const
  {
    for (size_t i = 0; i < m_elements.size(); ++i)
    {
      if (m_elements[i].Usage == usage && m_elements[i].UsageIndex == usageIndex)
        return i;
    }
    return -1;
  }


}
