#ifdef _WIN32
/****************************************************************************************************************************************************
* Copyright (c) 2014 Freescale Semiconductor, Inc.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*    * Redistributions of source code must retain the above copyright notice,
*      this list of conditions and the following disclaimer.
*
*    * Redistributions in binary form must reproduce the above copyright notice,
*      this list of conditions and the following disclaimer in the documentation
*      and/or other materials provided with the distribution.
*
*    * Neither the name of the Freescale Semiconductor, Inc. nor the names of
*      its contributors may be used to endorse or promote products derived from
*      this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
* IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
* BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
* OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
****************************************************************************************************************************************************/

#include <deque>
#include <iostream>
#include <FslBase/Math/Rectangle.hpp>
#include <FslBase/Log/Log.hpp>
#include <FslGraphics/Window/INativeWindowEventQueue.hpp>
#include <FslGraphicsEGL/Exceptions.hpp>
#include "EGLNativeWindowWin32.hpp"
#include <FslGraphics/Window/NativeWindowEventHelper.hpp>
#include <windows.h>
#include <windowsx.h>

namespace Fsl
{
  namespace
  {
    // For now we always force activation (meaning we dont lose deactivate on windows focus loss)
    bool g_forceActivated = true;

    std::weak_ptr<INativeWindowEventQueue> g_eventQueue;
    int g_activated = 0;
    bool g_isMouseCaptured;
    uint32_t g_mouseButtonState;

    bool TryConvert(WPARAM wParam, VirtualKey::Enum& rKey)
    {
      switch (wParam)
      {
      case VK_ESCAPE:
        rKey = VirtualKey::Escape;
        return true;
      case VK_TAB:
        rKey = VirtualKey::Tab;
        return true;
      case VK_RETURN:
        rKey = VirtualKey::Return;
        return true;
      case VK_SPACE:
        rKey = VirtualKey::Space;
        return true;
      case VK_BACK:
        rKey = VirtualKey::Backspace;
        return true;

      case VK_UP:
        rKey = VirtualKey::UpArrow;
        return true;
      case VK_DOWN:
        rKey = VirtualKey::DownArrow;
        return true;
      case VK_LEFT:
        rKey = VirtualKey::LeftArrow;
        return true;
      case VK_RIGHT:
        rKey = VirtualKey::RightArrow;
        return true;

      case VK_INSERT:
        rKey = VirtualKey::Insert;
        return true;
      case VK_DELETE:
        rKey = VirtualKey::Delete;
        return true;
      case VK_HOME:
        rKey = VirtualKey::Home;
        return true;
      case VK_END:
        rKey = VirtualKey::End;
        return true;
      case VK_PRIOR:
        rKey = VirtualKey::PageUp;
        return true;
      case VK_NEXT:
        rKey = VirtualKey::PageDown;
        return true;
      // Snapshot only sends a up event, so we can't really use it
      // There are workarounds, but its not worth using time on at the moment.
      //case VK_SNAPSHOT:
      //  rKey = VirtualKey::PrintScreen;
      //  return true;
      case VK_SCROLL:
        rKey = VirtualKey::ScrollLock;
        return true;
      case VK_PAUSE:
        rKey = VirtualKey::Pause;
        return true;
      case VK_CANCEL:
        rKey = VirtualKey::Break;
        return true;
      }

      if (wParam >= '0' && wParam <= '9')
      {
        // Take advantage of the simple encoding
        rKey = static_cast<VirtualKey::Enum>(VirtualKey::Code0 + (wParam - '0'));
        return true;
      }
      if (wParam >= 'A' && wParam <= 'Z')
      {
        // Take advantage of the simple encoding
        rKey = static_cast<VirtualKey::Enum>(VirtualKey::A + (wParam - 'A'));
        return true;
      }
      if (wParam >= VK_F1 && wParam <= VK_F12)
      {
        rKey = static_cast<VirtualKey::Enum>(VirtualKey::F1 + (wParam - VK_F1));
        return true;
      }

      return false;
    }


    LRESULT OnKeyMessage(const std::shared_ptr<INativeWindowEventQueue>& eventQueue, WPARAM wParam, LPARAM lParam, const bool isPressed)
    {
      VirtualKey::Enum keycode;
      if (TryConvert(wParam, keycode))
      {
        const NativeWindowEvent event = NativeWindowEventHelper::EncodeInputKeyEvent(keycode, isPressed);
        eventQueue->PostEvent(event);
      }
      return 0;
    }


    LRESULT OnMouseButton(HWND hWnd, const std::shared_ptr<INativeWindowEventQueue>& eventQueue, WPARAM wParam, LPARAM lParam, const VirtualMouseButton::Enum button, const bool isPressed)
    {
      const Point2 position(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
      const NativeWindowEvent event = NativeWindowEventHelper::EncodeInputMouseButtonEvent(button, isPressed, position);
      eventQueue->PostEvent(event);

      if (isPressed)
        g_mouseButtonState |= button;
      else
        g_mouseButtonState &= ~button;

      if (g_mouseButtonState != 0)
        SetCapture(hWnd);
      else
        ReleaseCapture();
      return 0;
    }


    LRESULT OnMouseMove(const std::shared_ptr<INativeWindowEventQueue>& eventQueue, WPARAM wParam, LPARAM lParam)
    {
      const Point2 position(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
      const NativeWindowEvent event = NativeWindowEventHelper::EncodeInputMouseMoveEvent(position);
      eventQueue->PostEvent(event);
      return 0;
    }


    LRESULT OnMouseWheel(const std::shared_ptr<INativeWindowEventQueue>& eventQueue, WPARAM wParam, LPARAM lParam)
    {
      const int32_t zDelta = GET_WHEEL_DELTA_WPARAM(wParam);
      const Point2 position(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
      const NativeWindowEvent event = NativeWindowEventHelper::EncodeInputMouseWheelEvent(zDelta, position);
      eventQueue->PostEvent(event);
      return 0;
    }

    LRESULT OnActivateMessage(const std::shared_ptr<INativeWindowEventQueue>& eventQueue, WPARAM wParam, LPARAM lParam)
    {
      const int hiWord = ((wParam >> 16) & 0xFFFF);
      const int lowWord = (wParam & 0xFFFF);

      // Order of commands on activate == false and minimize
      // - NativeWindowEventType::WindowActivation false
      // - NativeWindowEventType::WindowSuspend true
      // Order of commands on activate == true and restore
      // - NativeWindowEventType::WindowSuspend false
      // - NativeWindowEventType::WindowActivation true

      // When the window is minimized we send the suspend message
      //const bool isMinimize = hiWord != 0;
      //if (isMinimize && lowWord != 0)
      //  eventQueue->PostEvent(NativeWindowEventHelper::EncodeWindowSuspendEvent(false));

      // Ensure that we only send state changes
      const int activated = (lowWord != 0 || g_forceActivated ? 1 : 0);
      if (activated != g_activated)
      {
        g_activated = activated;
        eventQueue->PostEvent(NativeWindowEventHelper::EncodeWindowActivationEvent(activated != 0));
      }

      // When the window is minimized we send the suspend message
      //if (isMinimize && lowWord == 0)
      //  eventQueue->PostEvent(NativeWindowEventHelper::EncodeWindowSuspendEvent(true));
      return 0;
    }


    LRESULT OnSize(const std::shared_ptr<INativeWindowEventQueue>& eventQueue, WPARAM wParam, LPARAM lParam)
    {
      eventQueue->PostEvent(NativeWindowEventHelper::EncodeWindowResizedEvent());
      return 0;
    }


    LRESULT CALLBACK WndProc(HWND hWnd, UINT uiMsg, WPARAM wParam, LPARAM lParam)
    {
      std::shared_ptr<INativeWindowEventQueue> eventQueue = g_eventQueue.lock();

      if (uiMsg == WM_CLOSE)
      {
        PostQuitMessage(0);
        return 0;
      }
      else if (eventQueue)
      {
        switch (uiMsg)
        {
        case WM_ACTIVATE:
          return OnActivateMessage(eventQueue, wParam, lParam);
        case WM_KEYDOWN:
          return OnKeyMessage(eventQueue, wParam, lParam, true);
        case WM_KEYUP:
          return OnKeyMessage(eventQueue, wParam, lParam, false);
        case WM_LBUTTONDOWN:
          return OnMouseButton(hWnd, eventQueue, wParam, lParam, VirtualMouseButton::Left, true);
        case WM_LBUTTONUP:
          return OnMouseButton(hWnd, eventQueue, wParam, lParam, VirtualMouseButton::Left, false);
        case WM_MBUTTONDOWN:
          return OnMouseButton(hWnd, eventQueue, wParam, lParam, VirtualMouseButton::Middle, true);
        case WM_MBUTTONUP:
          return OnMouseButton(hWnd, eventQueue, wParam, lParam, VirtualMouseButton::Middle, false);
        case WM_RBUTTONDOWN:
          return OnMouseButton(hWnd, eventQueue, wParam, lParam, VirtualMouseButton::Right, true);
        case WM_RBUTTONUP:
          return OnMouseButton(hWnd, eventQueue, wParam, lParam, VirtualMouseButton::Right, false);
        case WM_MOUSEMOVE:
          return OnMouseMove(eventQueue, wParam, lParam);
        case WM_MOUSEWHEEL:
          return OnMouseWheel(eventQueue, wParam, lParam);
        case WM_SIZE:
          return OnSize(eventQueue, wParam, lParam);
        case WM_QUIT:
        case WM_QUERYENDSESSION:
        default:
          break;
        }
      }
      return DefWindowProc(hWnd, uiMsg, wParam, lParam);
    }


    //! @brief Adjust the client size of the window to the expected dimensions
    void SetClientSize(HWND hwnd, int clientWidth, int clientHeight)
    {
      if (IsWindow(hwnd))
      {
        DWORD dwStyle = GetWindowLongPtr(hwnd, GWL_STYLE);
        DWORD dwExStyle = GetWindowLongPtr(hwnd, GWL_EXSTYLE);
        HMENU menu = GetMenu(hwnd);

        RECT rc = { 0, 0, clientWidth, clientHeight };

        if (!AdjustWindowRectEx(&rc, dwStyle, menu ? TRUE : FALSE, dwExStyle))
          MessageBox(nullptr, TEXT("AdjustWindowRectEx Failed!"), TEXT("Error"), MB_OK);
        SetWindowPos(hwnd, nullptr, 0, 0, rc.right - rc.left, rc.bottom - rc.top, SWP_NOZORDER | SWP_NOMOVE);
      }
    }

    struct MonitorRecord
    {
      HMONITOR Handle;
      Rectangle Rect;

      MonitorRecord()
      {
      }

      MonitorRecord(HMONITOR hMonitor, Rectangle rect)
      {
        Handle = hMonitor;
        Rect = rect;
      }
    };

    std::deque<MonitorRecord> g_monitorDeque;

    BOOL CALLBACK MonitorEnumProc(_In_  HMONITOR hMonitor, _In_  HDC      hdcMonitor, _In_  LPRECT   lprcMonitor, _In_  LPARAM   dwData)
    {
      if (hdcMonitor == nullptr &&  lprcMonitor != nullptr)
      {
        const int32_t left = lprcMonitor->left;
        const int32_t top = lprcMonitor->top;
        const int32_t dx = lprcMonitor->right - left;
        const int32_t dy = lprcMonitor->bottom - top;
        if ( dx > 0 && dy > 0 )
          g_monitorDeque.push_back(MonitorRecord(hMonitor, Rectangle(left, top, dx, dy)));
      }
      return true;
    }

    bool TryGetMonitor(int32_t displayId, MonitorRecord& rRecord, int32_t& rMaxDisplayId)
    {
      rMaxDisplayId = 0;
      g_monitorDeque.clear();
      EnumDisplayMonitors(nullptr, nullptr, MonitorEnumProc, 0);
      rMaxDisplayId = static_cast<int32_t>(g_monitorDeque.size());
      if (displayId < 0 || displayId > rMaxDisplayId)
      {
        g_monitorDeque.clear();
        rRecord = MonitorRecord();
        return false;
      }
      rRecord = g_monitorDeque[displayId];
      g_monitorDeque.clear();
      return true;
    }

    MonitorRecord GetMonitorRecord(const int32_t requestedDisplayId)
    {
      int32_t maxDisplayId;
      MonitorRecord record;
      if (!TryGetMonitor(requestedDisplayId, record, maxDisplayId))
      {
        FSLLOG_WARNING("Backend max display Id is: " << maxDisplayId << " so " << requestedDisplayId << " is unsupported, using displayid 0.");
        if (!TryGetMonitor(0, record, maxDisplayId))
          throw EGLGraphicsException("Could not find a valid monitor");
      }
      return record;
    }
  }


  EGLNativeWindowSystemWin32::EGLNativeWindowSystemWin32(const NativeWindowSystemSetup& setup)
  {
    //SetProcessDPIAware();
  }


  EGLNativeWindowSystemWin32::~EGLNativeWindowSystemWin32()
  {
  }


  EGLNativeDisplayType EGLNativeWindowSystemWin32::GetDisplayType() const
  {
    return EGL_DEFAULT_DISPLAY;
  }


  std::shared_ptr<IEGLNativeWindow> EGLNativeWindowSystemWin32::CreateNativeWindow(const NativeWindowSetup& nativeWindowSetup, const NativeEGLSetup& nativeEglSetup)
  {
    return std::shared_ptr<IEGLNativeWindow>(new EGLNativeWindowWin32(nativeWindowSetup, nativeEglSetup));
  }


  bool EGLNativeWindowSystemWin32::ProcessMessages(const NativeWindowProcessMessagesArgs& args)
  {
    MSG sMessage;
    bool bQuit = false;
    if (PeekMessage(&sMessage, nullptr, 0, 0, PM_REMOVE))
    {
      switch (sMessage.message)
      {
      case WM_QUIT:
        bQuit = true;
        break;
      default:
        break;
      }

      TranslateMessage(&sMessage);
      DispatchMessage(&sMessage);
    }
    return !bQuit;
  }


  EGLNativeWindowWin32::EGLNativeWindowWin32(const NativeWindowSetup& nativeWindowSetup, const NativeEGLSetup& nativeEglSetup)
    : m_hWindow(nullptr)
  {
    g_eventQueue = nativeWindowSetup.GetEventQueue();

    const NativeWindowConfig nativeWindowConfig = nativeWindowSetup.GetConfig();
    WNDCLASS wc;
    RECT rect;
    HINSTANCE hInstance;

    Rectangle targetRectangle = nativeWindowConfig.GetWindowRectangle();

    bool bFullscreen = false;
    DWORD dwStyle = WS_OVERLAPPEDWINDOW | WS_CLIPSIBLINGS | WS_CLIPCHILDREN;
    switch (nativeWindowConfig.GetWindowMode())
    {
    case WindowMode::Fullscreen:
      {
        const int32_t displayIdRequest = nativeWindowSetup.GetConfig().GetDisplayId();
        if (displayIdRequest == 0)
        {
          const int width = GetSystemMetrics(SM_CXSCREEN);
          const int height = GetSystemMetrics(SM_CYSCREEN);
          targetRectangle = Rectangle(0, 0, width, height);
        }
        else
        {
          const MonitorRecord record = GetMonitorRecord(displayIdRequest);
          targetRectangle = record.Rect;
        }

        dwStyle = WS_POPUP | WS_VISIBLE | WS_SYSMENU;
        bFullscreen = true;
      }
      break;
    case WindowMode::Window:
      break;
    default:
      FSLLOG("WARNING: Unknown window mode");
      break;
    }


    rect.left = targetRectangle.Left();
    rect.right = targetRectangle.Right();
    rect.top = targetRectangle.Top();
    rect.bottom = targetRectangle.Bottom();

    hInstance = GetModuleHandle(nullptr);

    wc.style = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
    wc.lpfnWndProc = (WNDPROC)WndProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = 0;
    wc.hInstance = hInstance;
    wc.hIcon = LoadIcon(nullptr, IDI_WINLOGO);
    wc.hCursor = LoadCursor(nullptr, IDC_ARROW);
    wc.hbrBackground = nullptr;
    wc.lpszMenuName = nullptr;
    wc.lpszClassName = TEXT("OpenGLES");

    RegisterClass(&wc);

    m_hWindow = CreateWindowEx(WS_EX_APPWINDOW | WS_EX_WINDOWEDGE, TEXT("OpenGLES"), TEXT("Main"), dwStyle, targetRectangle.Left(), targetRectangle.Top(), targetRectangle.Width, targetRectangle.Height, nullptr, nullptr, hInstance, nullptr);
    if (m_hWindow == nullptr)
      throw EGLGraphicsException("Failed to create native window");

    //if (bFullscreen)
    //  SetWindowLong(m_hWindow, GWL_STYLE, 0);

    SetClientSize(m_hWindow, targetRectangle.Width, targetRectangle.Height);

    ShowWindow(m_hWindow, SW_SHOW);
    SetForegroundWindow(m_hWindow);
    SetFocus(m_hWindow);
  }


  EGLNativeWindowWin32::~EGLNativeWindowWin32()
  {
    if (m_hWindow != nullptr)
    {
      DestroyWindow(m_hWindow);
      m_hWindow = nullptr;
    }
  }


  EGLNativeWindowType EGLNativeWindowWin32::GetWindowType() const
  {
    return m_hWindow;
  }

}
#endif
