/****************************************************************************************************************************************************
* Copyright (c) 2014 Freescale Semiconductor, Inc.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*    * Redistributions of source code must retain the above copyright notice,
*      this list of conditions and the following disclaimer.
*
*    * Redistributions in binary form must reproduce the above copyright notice,
*      this list of conditions and the following disclaimer in the documentation
*      and/or other materials provided with the distribution.
*
*    * Neither the name of the Freescale Semiconductor, Inc. nor the names of
*      its contributors may be used to endorse or promote products derived from
*      this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
* IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
* BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
* OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
****************************************************************************************************************************************************/

#include <FslGraphics/Bitmap.hpp>
#include <FslGraphicsGLES2/Exceptions.hpp>
#include <FslGraphicsGLES2/GLCheck.hpp>
#include <FslGraphicsGLES2/GLRawBitmapUtil.hpp>
#include <FslGraphicsGLES2/GLTexture.hpp>

#include <algorithm>
#include <cassert>

namespace Fsl
{
  namespace GLES2
  {

    GLTexture::GLTexture()
      : m_handle(GLValues::INVALID_HANDLE)
      , m_target(GL_TEXTURE_2D)
      , m_size()
    {
    }


    GLTexture::GLTexture(const GLuint handle, const Point2& size)
      : m_handle(handle)
      , m_target(GL_TEXTURE_2D)
      , m_size(size)
    {
    }

    GLTexture::GLTexture(const GLuint handle, const GLenum target, const Point2& size)
      : m_handle(handle)
      , m_target(target)
      , m_size(size)
    {
      if (target != GL_TEXTURE_2D && target != GL_TEXTURE_CUBE_MAP)
        throw std::invalid_argument("Unsupported texture target type");
    }


    GLTexture::GLTexture(const Bitmap& bitmap, const GLTextureParameters& textureParameters)
      : m_handle(GLValues::INVALID_HANDLE)
      , m_target(GL_TEXTURE_2D)
      , m_size()
    {
      Reset(bitmap, textureParameters);
    }


    GLTexture::GLTexture(const RawBitmap& bitmap, const GLTextureParameters& textureParameters)
      : m_handle(GLValues::INVALID_HANDLE)
      , m_target(GL_TEXTURE_2D)
      , m_size()
    {
      Reset(bitmap, textureParameters);
    }


    GLTexture::GLTexture(const Bitmap& bitmap, const GLTextureParameters& textureParameters, const bool generateMipMaps)
      : m_handle(GLValues::INVALID_HANDLE)
      , m_target(GL_TEXTURE_2D)
      , m_size()
    {
      Reset(bitmap, textureParameters, generateMipMaps);
    }


    GLTexture::GLTexture(const RawBitmap& bitmap, const GLTextureParameters& textureParameters, const bool generateMipMaps)
      : m_handle(GLValues::INVALID_HANDLE)
      , m_target(GL_TEXTURE_2D)
      , m_size()
    {
      Reset(bitmap, textureParameters, generateMipMaps);
    }


    GLTexture::GLTexture(const Bitmap& bitmapPosX, const Bitmap& bitmapNegX, const Bitmap& bitmapPosY, const Bitmap& bitmapNegY, const Bitmap& bitmapPosZ, const Bitmap& bitmapNegZ, const GLTextureParameters& textureParameters)
      : m_handle(GLValues::INVALID_HANDLE)
      , m_target(GL_TEXTURE_2D)
      , m_size()
    {
      Reset(bitmapPosX, bitmapNegX, bitmapPosY, bitmapNegY, bitmapPosZ, bitmapNegZ, textureParameters);
    }


    GLTexture::GLTexture(const RawBitmap& bitmapPosX, const RawBitmap& bitmapNegX, const RawBitmap& bitmapPosY, const RawBitmap& bitmapNegY, const RawBitmap& bitmapPosZ, const RawBitmap& bitmapNegZ, const GLTextureParameters& textureParameters)
      : m_handle(GLValues::INVALID_HANDLE)
      , m_target(GL_TEXTURE_2D)
      , m_size()
    {
      Reset(RawCubeBitmap(bitmapPosX, bitmapNegX, bitmapPosY, bitmapNegY, bitmapPosZ, bitmapNegZ), textureParameters);
    }


    GLTexture::GLTexture(const RawCubeBitmap& cubeBitmap, const GLTextureParameters& textureParameters)
      : m_handle(GLValues::INVALID_HANDLE)
      , m_target(GL_TEXTURE_2D)
      , m_size()
    {
      Reset(cubeBitmap, textureParameters);
    }


    GLTexture::GLTexture(const Bitmap& bitmapPosX, const Bitmap& bitmapNegX, const Bitmap& bitmapPosY, const Bitmap& bitmapNegY, const Bitmap& bitmapPosZ, const Bitmap& bitmapNegZ, const GLTextureParameters& textureParameters, const bool generateMipMaps)
      : m_handle(GLValues::INVALID_HANDLE)
      , m_target(GL_TEXTURE_2D)
      , m_size()
    {
      Reset(bitmapPosX, bitmapNegX, bitmapPosY, bitmapNegY, bitmapPosZ, bitmapNegZ, textureParameters, generateMipMaps);
    }


    GLTexture::GLTexture(const RawBitmap& bitmapPosX, const RawBitmap& bitmapNegX, const RawBitmap& bitmapPosY, const RawBitmap& bitmapNegY, const RawBitmap& bitmapPosZ, const RawBitmap& bitmapNegZ, const GLTextureParameters& textureParameters, const bool generateMipMaps)
      : m_handle(GLValues::INVALID_HANDLE)
      , m_target(GL_TEXTURE_2D)
      , m_size()
    {
      Reset(RawCubeBitmap(bitmapPosX, bitmapNegX, bitmapPosY, bitmapNegY, bitmapPosZ, bitmapNegZ), textureParameters, generateMipMaps);
    }


    GLTexture::GLTexture(const RawCubeBitmap& cubeBitmap, const GLTextureParameters& textureParameters, const bool generateMipMaps)
      : m_handle(GLValues::INVALID_HANDLE)
      , m_target(GL_TEXTURE_2D)
      , m_size()
    {
      Reset(cubeBitmap, textureParameters, generateMipMaps);
    }


    GLTexture::~GLTexture()
    {
      Reset();
    }


    void GLTexture::Reset()
    {
      if (m_handle != GLValues::INVALID_HANDLE)
      {
        glDeleteTextures(1, &m_handle);
        m_handle = GLValues::INVALID_HANDLE;
        m_target = GL_TEXTURE_2D;
        m_size = Point2();
      }
    }


    void GLTexture::Reset(const GLuint handle, const Point2& size)
    {
      Reset();
      m_handle = handle;
      m_target = GL_TEXTURE_2D;
      m_size = size;
    }


    void GLTexture::Reset(const GLuint handle, const GLenum target, const Point2& size)
    {
      if (target != GL_TEXTURE_2D && target != GL_TEXTURE_CUBE_MAP)
        throw std::invalid_argument("Unsupported texture target type");

      Reset();
      m_handle = handle;
      m_target = target;
      m_size = size;
    }


    void GLTexture::Reset(const Bitmap& bitmap, const GLTextureParameters& textureParameters)
    {
      Reset();
      SetData(bitmap, textureParameters);
    }


    void GLTexture::Reset(const RawBitmap& bitmap, const GLTextureParameters& textureParameters)
    {
      Reset();
      SetData(bitmap, textureParameters);
    }


    void GLTexture::Reset(const Bitmap& bitmap, const GLTextureParameters& textureParameters, const bool generateMipMaps)
    {
      Reset();
      SetData(bitmap, textureParameters, generateMipMaps);
    }


    void GLTexture::Reset(const RawBitmap& bitmap, const GLTextureParameters& textureParameters, const bool generateMipMaps)
    {
      Reset();
      SetData(bitmap, textureParameters, generateMipMaps);
    }


    void GLTexture::Reset(const Bitmap& bitmapPosX, const Bitmap& bitmapNegX, const Bitmap& bitmapPosY, const Bitmap& bitmapNegY, const Bitmap& bitmapPosZ, const Bitmap& bitmapNegZ, const GLTextureParameters& textureParameters)
    {
      Reset();
      SetData(bitmapPosX, bitmapNegX, bitmapPosY, bitmapNegY, bitmapPosZ, bitmapNegZ, textureParameters);
    }


    void GLTexture::Reset(const RawBitmap& bitmapPosX, const RawBitmap& bitmapNegX, const RawBitmap& bitmapPosY, const RawBitmap& bitmapNegY, const RawBitmap& bitmapPosZ, const RawBitmap& bitmapNegZ, const GLTextureParameters& textureParameters)
    {
      Reset();
      SetData(RawCubeBitmap(bitmapPosX, bitmapNegX, bitmapPosY, bitmapNegY, bitmapPosZ, bitmapNegZ), textureParameters);
    }


    void GLTexture::Reset(const RawCubeBitmap& cubeBitmap, const GLTextureParameters& textureParameters)
    {
      Reset();
      SetData(cubeBitmap, textureParameters);
    }


    void GLTexture::Reset(const Bitmap& bitmapPosX, const Bitmap& bitmapNegX, const Bitmap& bitmapPosY, const Bitmap& bitmapNegY, const Bitmap& bitmapPosZ, const Bitmap& bitmapNegZ, const GLTextureParameters& textureParameters, const bool generateMipMaps)
    {
      Reset();
      SetData(bitmapPosX, bitmapNegX, bitmapPosY, bitmapNegY, bitmapPosZ, bitmapNegZ, textureParameters, generateMipMaps);
    }


    void GLTexture::Reset(const RawBitmap& bitmapPosX, const RawBitmap& bitmapNegX, const RawBitmap& bitmapPosY, const RawBitmap& bitmapNegY, const RawBitmap& bitmapPosZ, const RawBitmap& bitmapNegZ, const GLTextureParameters& textureParameters, const bool generateMipMaps)
    {
      Reset();
      SetData(RawCubeBitmap(bitmapPosX, bitmapNegX, bitmapPosY, bitmapNegY, bitmapPosZ, bitmapNegZ), textureParameters, generateMipMaps);
    }


    void GLTexture::Reset(const RawCubeBitmap& cubeBitmap, const GLTextureParameters& textureParameters, const bool generateMipMaps)
    {
      Reset();
      SetData(cubeBitmap, textureParameters, generateMipMaps);
    }


    void GLTexture::SetData(const Bitmap& bitmap, const GLTextureParameters& textureParameters)
    {
      RawBitmap rawBitmap;
      Bitmap::ScopedDirectAccess scopedAccess(bitmap, rawBitmap);
      SetData(rawBitmap, textureParameters, false);
    }


    void GLTexture::SetData(const RawBitmap& bitmap, const GLTextureParameters& textureParameters)
    {
      SetData(bitmap, textureParameters, false);
    }


    void GLTexture::SetData(const RawCubeBitmap& cubeBitmap, const GLTextureParameters& textureParameters)
    {
      SetData(cubeBitmap, textureParameters, false);
    }


    void GLTexture::SetData(const Bitmap& bitmap, const GLTextureParameters& textureParameters, const bool generateMipMaps)
    {
      RawBitmap rawBitmap;
      Bitmap::ScopedDirectAccess scopedAccess(bitmap, rawBitmap);
      SetData(rawBitmap, textureParameters, generateMipMaps);
    }


    void GLTexture::SetData(const RawBitmap& bitmap, const GLTextureParameters& textureParameters, const bool generateMipMaps)
    {
      if (!bitmap.IsValid())
        throw std::invalid_argument("The bitmaps must be valid");
      if (!textureParameters.IsValid())
        throw std::invalid_argument("The texture parameters are invalid");

      const GLRawBitmapUtil::Result result = GLRawBitmapUtil::Convert(bitmap);

      if (m_target != GL_TEXTURE_2D)
      {
        Reset();
      }

      if (m_handle == GLValues::INVALID_HANDLE)
      {
        GL_CHECK(glGenTextures(1, &m_handle));
      }

      GL_CHECK(glBindTexture(GL_TEXTURE_2D, m_handle));

      GL_CHECK(glPixelStorei(GL_UNPACK_ALIGNMENT, result.Alignment));

      GL_CHECK(glTexImage2D(GL_TEXTURE_2D, 0, result.Format, bitmap.Width(), bitmap.Height(), 0, result.Format, result.Type, bitmap.Content()));

      if (generateMipMaps)
      {
        GL_CHECK(glGenerateMipmap(GL_TEXTURE_2D));
      }

      GL_CHECK(glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, textureParameters.MinFilter));
      GL_CHECK(glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, textureParameters.MagFilter));
      GL_CHECK(glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, textureParameters.WrapS));
      GL_CHECK(glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, textureParameters.WrapT));

      m_size = Point2(bitmap.Width(), bitmap.Height());
      m_target = GL_TEXTURE_2D;
    }


    void GLTexture::SetData(const Bitmap& bitmapPosX, const Bitmap& bitmapNegX, const Bitmap& bitmapPosY, const Bitmap& bitmapNegY, const Bitmap& bitmapPosZ, const Bitmap& bitmapNegZ, const GLTextureParameters& textureParameters)
    {
      RawBitmap rawBitmapPosX, rawBitmapNegX, rawBitmapPosY, rawBitmapNegY, rawBitmapPosZ, rawBitmapNegZ;
      Bitmap::ScopedDirectAccess scopedAccessXP(bitmapPosX, rawBitmapPosX);
      Bitmap::ScopedDirectAccess scopedAccessXN(bitmapNegX, rawBitmapNegX);
      Bitmap::ScopedDirectAccess scopedAccessYP(bitmapPosY, rawBitmapPosY);
      Bitmap::ScopedDirectAccess scopedAccessYN(bitmapNegY, rawBitmapNegY);
      Bitmap::ScopedDirectAccess scopedAccessZP(bitmapPosZ, rawBitmapPosZ);
      Bitmap::ScopedDirectAccess scopedAccessZN(bitmapNegZ, rawBitmapNegZ);
      SetData(RawCubeBitmap(rawBitmapPosX, rawBitmapNegX, rawBitmapPosY, rawBitmapNegY, rawBitmapPosZ, rawBitmapNegZ), textureParameters, false);
    }


    void GLTexture::SetData(const RawBitmap& bitmapPosX, const RawBitmap& bitmapNegX, const RawBitmap& bitmapPosY, const RawBitmap& bitmapNegY, const RawBitmap& bitmapPosZ, const RawBitmap& bitmapNegZ, const GLTextureParameters& textureParameters)
    {
      return SetData(bitmapPosX, bitmapNegX, bitmapPosY, bitmapNegY, bitmapPosZ, bitmapNegZ, textureParameters, false);
    }


    void GLTexture::SetData(const Bitmap& bitmapPosX, const Bitmap& bitmapNegX, const Bitmap& bitmapPosY, const Bitmap& bitmapNegY, const Bitmap& bitmapPosZ, const Bitmap& bitmapNegZ, const GLTextureParameters& textureParameters, const bool generateMipMaps)
    {
      RawBitmap rawBitmapPosX, rawBitmapNegX, rawBitmapPosY, rawBitmapNegY, rawBitmapPosZ, rawBitmapNegZ;
      Bitmap::ScopedDirectAccess scopedAccessXP(bitmapPosX, rawBitmapPosX);
      Bitmap::ScopedDirectAccess scopedAccessXN(bitmapNegX, rawBitmapNegX);
      Bitmap::ScopedDirectAccess scopedAccessYP(bitmapPosY, rawBitmapPosY);
      Bitmap::ScopedDirectAccess scopedAccessYN(bitmapNegY, rawBitmapNegY);
      Bitmap::ScopedDirectAccess scopedAccessZP(bitmapPosZ, rawBitmapPosZ);
      Bitmap::ScopedDirectAccess scopedAccessZN(bitmapNegZ, rawBitmapNegZ);
      SetData(RawCubeBitmap(rawBitmapPosX, rawBitmapNegX, rawBitmapPosY, rawBitmapNegY, rawBitmapPosZ, rawBitmapNegZ), textureParameters, generateMipMaps);
    }


    void GLTexture::SetData(const RawBitmap& bitmapPosX, const RawBitmap& bitmapNegX, const RawBitmap& bitmapPosY, const RawBitmap& bitmapNegY, const RawBitmap& bitmapPosZ, const RawBitmap& bitmapNegZ, const GLTextureParameters& textureParameters, const bool generateMipMaps)
    {
      SetData(RawCubeBitmap(bitmapPosX, bitmapNegX, bitmapPosY, bitmapNegY, bitmapPosZ, bitmapNegZ), textureParameters, generateMipMaps);
    }


    void GLTexture::SetData(const RawCubeBitmap& cubeBitmap, const GLTextureParameters& textureParameters, const bool generateMipMaps)
    {
      if (!cubeBitmap.IsValid())
        throw std::invalid_argument("The cube bitmap must be valid");

      if (!textureParameters.IsValid())
        throw std::invalid_argument("The texture parameters are invalid");

      const GLRawBitmapUtil::Result resultPosX = GLRawBitmapUtil::Convert(cubeBitmap.GetPosX());
      const GLRawBitmapUtil::Result resultNegX = GLRawBitmapUtil::Convert(cubeBitmap.GetNegX());
      const GLRawBitmapUtil::Result resultPosY = GLRawBitmapUtil::Convert(cubeBitmap.GetPosY());
      const GLRawBitmapUtil::Result resultNegY = GLRawBitmapUtil::Convert(cubeBitmap.GetNegY());
      const GLRawBitmapUtil::Result resultPosZ = GLRawBitmapUtil::Convert(cubeBitmap.GetPosZ());
      const GLRawBitmapUtil::Result resultNegZ = GLRawBitmapUtil::Convert(cubeBitmap.GetNegZ());


      const GLint texFormat = resultPosX.Format;

      if (m_target != GL_TEXTURE_CUBE_MAP)
      {
        Reset();
      }

      if (m_handle == GLValues::INVALID_HANDLE)
      {
        GL_CHECK(glGenTextures(1, &m_handle));
      }

      const int texWidth = cubeBitmap.Width();
      const int texHeight = cubeBitmap.Height();

      GL_CHECK(glBindTexture(GL_TEXTURE_CUBE_MAP, m_handle));
      GL_CHECK(glPixelStorei(GL_UNPACK_ALIGNMENT, resultPosX.Alignment));
      GL_CHECK(glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X, 0, texFormat, texWidth, texHeight, 0, texFormat, resultPosX.Type, cubeBitmap.GetPosX().Content()));
      GL_CHECK(glPixelStorei(GL_UNPACK_ALIGNMENT, resultNegX.Alignment));
      GL_CHECK(glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_X, 0, texFormat, texWidth, texHeight, 0, texFormat, resultNegX.Type, cubeBitmap.GetNegX().Content()));
      GL_CHECK(glPixelStorei(GL_UNPACK_ALIGNMENT, resultPosY.Alignment));
      GL_CHECK(glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_Y, 0, texFormat, texWidth, texHeight, 0, texFormat, resultPosY.Type, cubeBitmap.GetPosY().Content()));
      GL_CHECK(glPixelStorei(GL_UNPACK_ALIGNMENT, resultNegY.Alignment));
      GL_CHECK(glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_Y, 0, texFormat, texWidth, texHeight, 0, texFormat, resultNegY.Type, cubeBitmap.GetNegY().Content()));
      GL_CHECK(glPixelStorei(GL_UNPACK_ALIGNMENT, resultPosZ.Alignment));
      GL_CHECK(glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_Z, 0, texFormat, texWidth, texHeight, 0, texFormat, resultPosZ.Type, cubeBitmap.GetPosZ().Content()));
      GL_CHECK(glPixelStorei(GL_UNPACK_ALIGNMENT, resultNegZ.Alignment));
      GL_CHECK(glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_Z, 0, texFormat, texWidth, texHeight, 0, texFormat, resultNegZ.Type, cubeBitmap.GetNegZ().Content()));

      if (generateMipMaps)
      {
        GL_CHECK(glGenerateMipmap(GL_TEXTURE_CUBE_MAP));
      }

      GL_CHECK(glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, textureParameters.MinFilter));
      GL_CHECK(glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, textureParameters.MagFilter));
      GL_CHECK(glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_S, textureParameters.WrapS));
      GL_CHECK(glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_T, textureParameters.WrapT));

      m_size = Point2(texWidth, texHeight);
      m_target = GL_TEXTURE_CUBE_MAP;
    }


    GLuint GLTexture::GetHandle() const
    {
      return m_handle;
    }


    GLuint GLTexture::GetTarget() const
    {
      return m_target;
    }


    Point2 GLTexture::GetSize() const
    {
      return m_size;
    }


    NativeTextureArea GLTexture::CalcTextureArea(const TextureRectangle& textureRectangle, const int timesX, const int timesY)
    {
      const Rectangle srcRect = textureRectangle.GetRectangle();
      const Point2 sizeTex = textureRectangle.GetTextureSize();

      assert(timesX == 1 || (timesX != 1 && (srcRect.X == 0 || srcRect.Width == sizeTex.X)));
      assert(timesY == 1 || (timesY != 1 && (srcRect.Y == 0 || srcRect.Height == sizeTex.Y)));

      assert(static_cast<float>(sizeTex.X) >= 0.0f);
      assert(static_cast<float>(sizeTex.Y) >= 0.0f);
      return NativeTextureArea(srcRect.Left() == 0 ? 0.0f : srcRect.Left() / static_cast<float>(sizeTex.X),
        srcRect.Top() == 0 ? 0.0f : srcRect.Top() / static_cast<float>(sizeTex.Y),
        srcRect.Right() == sizeTex.X ? 1.0f * timesX : srcRect.Right() / static_cast<float>(sizeTex.X),
        srcRect.Bottom() == sizeTex.Y ? 1.0f * timesY : srcRect.Bottom() / static_cast<float>(sizeTex.Y));
    }
  }

}
